/*
	$Id: UploadedFile.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.fileupload;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
	Objects of this class can be used to get information about
	an uploaded file and save the file to a directory in the filesystem.
	If the file is not saved, it will be deleted once the object is garbage
	collected.
	
	@author Nicklas
	@version 2.0
*/
public class UploadedFile
{
	private FileUpload fu;
	private final String formName;
	private final String filename;
	private final String mimeType;

	/**
		Create a new <code>UploadedFile</code> object. 
		Use the {@link FileUpload#getFile(String)} or {@link
		FileUpload#getFileValues(String)}
		method to get an object representing the uploaded file.
		@param fu The <code>FileUpload</code> object responsible for the upload
		@param formName The name of the form field with this file
		@param mimeType The MIME type of the file, or NULL if it is not known
		@param filename The original filename of the file, or NULL if it is not known
	*/
	UploadedFile(FileUpload fu, String formName, String filename, String mimeType)
	{
		this.fu = fu;
		this.formName = formName;
		this.filename = filename;
		this.mimeType = mimeType;
	}

	/**
		Get the name of the form field that this file was selected in.
		@since 2.16
	*/
	public String getFormName()
	{
		return formName;
	}
	
	/**
		Get the original filename of the uploaded file. This is the name the file had
		on the client computer.
		@return The original name of the file, or NULL if it is not known
	*/
	public String getFilename()
	{
		return filename;
	}

	/**
		Get the extension of the original filename. The extension is the part of the filename
		after the last dot. If the file doesn't have an extension NULL is returned.
		@return The extension of the file, or NULL if none exists
	*/
	public String getExtension()
	{
		int dot = filename.lastIndexOf('.');
		if (dot >= 0)
		{
			return filename.substring(dot+1);
		}
		else
		{
			return null;
		}
	}

	/**
		Get the MIME type of the uploaded file.
		@return The MIME type of the file, or NULL if it is not known
	*/
	public String getMimeType()
	{
		return mimeType;
	}

	/**
		Upload the file and write the data to the specified output stream.
		@param out The stream to write to

	*/
	public void writeTo(OutputStream out)
		throws IOException, UploadAbortedException
	{
		fu.doUpload(out, this);
	}
	
	/**
		Get an input stream for uploading the file.
		@return An input stream to read from
	 	@throws IOException If reading the input stream fails. 
	 	@throws UploadAbortedException If the upload has been requested to be aborted.
	*/
	public InputStream getInputStream()
		throws IOException, UploadAbortedException
	{
		return fu.getInputStream(this);
	}

}


