/*
	$Id: TabControl.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.tab;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import net.sf.basedb.clients.web.extensions.DynamicActionAttributeSupport;
import net.sf.basedb.clients.web.extensions.tabcontrol.TabAction;
import net.sf.basedb.clients.web.taglib.StylableTag;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Renderer;

/**
	This is the main tag for the tab control. It must always be present as the
	outermost tag. Use one or more {@link Tab} tags to define each tab inside the
	tabcontrol.
	<p>
	<b>Syntax:</b>
	<pre>
   &lt;t:tabcontrol
      id=...
      clazz=...
      subclass=...
      style=...
      contentstyle=...
      active=...
      position=...
      notabs=true|false
      extensions=...
   &gt;
</pre>


	<table border=1 cellspacing=0 cellpadding=2>
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>yes</td>
		<td>The ID of the tab control. The value of this attribute
		goes directly into the standard <code>id</code> attribute.
		This means that if you need a reference to the tab control object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		tabcontrol = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>tabcontrol</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>subclass</td>
		<td>-</td>
		<td>no</td>
		<td>
			Additional classes to put in the HTML class attribute. Doesn't
			replace the main class attribute.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tabcontrols.
		</td>
	</tr>
	<tr>
		<td>contentstyle</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute for the content section. This allows you to apply 
			different styles to different tabcontrols.
		</td>
	</tr>
	<tr>
		<td>active</td>
		<td>-</td>
		<td>no</td>
		<td>
			The id of the tab that should be active to begin with. If no value is
			specified the first tab is automatically made active.
		</td>
	</tr>
	<tr>
		<td>position</td>
		<td>top</td>
		<td>no</td>
		<td>The position of the tabs in relation to the content. The default value is to display
		the tabs at the top. The only other allowed value is <code>bottom</code>.
		</td>
	</tr>
	<tr>
		<td>notabs</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the tabcontrol should display tabs or not. If false the tab control will
			not output any HTML except the contents of the active tab.
		</td>
	</tr>
	<tr>
		<td>extensions</td>
		<td>-</td>
		<td>no</td>
		<td>
			An {@link ExtensionsInvoker} for {@link TabAction} extensions. All
			actions will be rendered by this tabcontrol after all	the 
			normally defined tags have been rendered. Since BASE 2.17.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
	@see Tab
*/
public class TabControl
	extends StylableTag
	implements Renderer<TabAction>
{
	/**
		
	*/
	private static final long serialVersionUID = 1898616270560551262L;

	/**
		Optional <code>style</code> attribute for the content section
	*/
	private String contentStyle = null;
	
	/**
		The ID attribute of the active tab.
	*/
	private String active = null;
	
	/**
		The position of the tabs.
	*/
	private int position = TOP;
	
	/**
	 	If it should try to remember the last active tab on the page
	*/
	private boolean remember = true;
	
	/**
		If the tabs should be displayed or not.
	*/
	private boolean noTabs = false;
	
	private ExtensionsInvoker<TabAction> extensions;
	
	/**
		The ID of the tab that is initially active.
	*/
	private String initialTab = null;
	
	/**
		The number of tabs.
	*/
	private int numTabs = 0;
	
	/**
		The HTML for the tabs.
	*/
	private StringBuilder tabs;
	
	/**
		The HTML for the contents of the tabs.
	*/
	private StringBuilder content;
	
	/**
		Tabs are at the top of the tab control.
	*/
	public static final int TOP = 1;
	
	/**
		Tabs are at the bottom of the tab control.
	*/
	public static final int BOTTOM = 2;

	public TabControl()
	{
		super("tabcontrol");
	}
	
	/*
		Taglib initialization methods
	*/
	public void setContentstyle(String contentStyle)
	{
		this.contentStyle = contentStyle;
	}
	public String getContentstyle()
	{
		return contentStyle;
	}

	public void setActive(String active)
	{
		this.active = active;
	}
	public String getActive()
	{
		return active;
	}
	
	public void setPosition(String position)
	{
		if ("bottom".equalsIgnoreCase(position))
		{
			this.position = BOTTOM;
		}
		else
		{
			this.position = TOP;
		}
	}

	public void setExtensions(ExtensionsInvoker<TabAction> extensions)
	{
		this.extensions = extensions;
	}

	
	public void setRemember(boolean remember)
	{
		this.remember = remember;
	}
	
	public void setNotabs(boolean noTabs)
	{
		this.noTabs = noTabs;
	}

	void addTab(TabAction tab)
	{
		numTabs++;
		if ((active != null) && active.equals(tab.getId()))
		{
			initialTab = tab.getId();
		}

		if (!noTabs)
		{
			String className = "tab interactable";
			if (tab.getClazz() != null) className += " " + tab.getClazz();
	
			tabs.append("<td id=\"").append(getId()).append(".").append(tab.getId()).append("\"");
			tabs.append(" class=\"").append(className).append(" auto-init\" data-auto-init=\"tab\"");
			tabs.append(DynamicActionAttributeSupport.getAttributesString(tab));
			if (tab.getHelpId() != null)
			{
				tabs.append(" data-help-id=\"").append(tab.getHelpId()).append("\"");
			}
			if (tab.getTabIndex() >= 0) 
			{
				tabs.append(" tabindex=\"").append(tab.getTabIndex()).append("\"");
			}
			if (tab.getTooltip() != null) tabs.append(" title=\"").append(tab.getTooltip()).append("\"");
			tabs.append(">").append(tab.getTitle()).append("</td><td class=\"tabspace\"></td>\n");
		}
		
		String tabContent = tab.getContent();
		if (tabContent != null)
		{
			content.append("<div id=\"").append(getId()).append(".").append(tab.getId()).append(".content\"");
			if (!noTabs || getContentstyle() != null)
			{
				content.append(" style=\"");
	 			if (!noTabs) content.append("display:none;");
	 			if (getContentstyle() != null) content.append(getContentstyle());
	 			content.append("\"");
			}
 			content.append(">");
			content.append(tabContent);
			content.append("</div>\n");
		}

	}

	@Override
	public int doStartTag()
		throws JspException
	{
		tabs = new StringBuilder();
		content = new StringBuilder();
		initialTab = null;

		if (!noTabs)
		{
			tabs.append("<table class=\"tabscontainer\">");
			tabs.append("<tr>\n");
			tabs.append("<td class=\"tabfillbefore\"></td>\n");
			tabs.append("<td class=\"tabs\">\n");
			tabs.append("<table>\n");
			tabs.append("<tr>\n");
		}

		return EVAL_BODY_INCLUDE;
	}

	@Override
	public int doEndTag() 
		throws JspException
	{
		if (extensions != null) extensions.render(this);
		
		StringBuilder sb = new StringBuilder();
		sb.append("<div ");
		addIdAndStyles(sb, "auto-init", noTabs ? "notabs" : (position == BOTTOM ? "tabsatbottom" : "tabsattop"));
		sb.append(" data-auto-init=\"tabcontrol\"");
		if (initialTab != null) 
		{
			sb.append(" data-initial-tab=\"").append(initialTab).append("\"");
		}
		sb.append(" data-remember-last=\"").append(remember ? "1" : "0").append("\"");
		addDynamicAttributes(sb);
		sb.append(">\n");

		if (!noTabs)
		{
			tabs.append("</tr>\n");
			tabs.append("</table>\n");
			tabs.append("</td>\n");
			tabs.append("<td class=\"tabfillafter\"></td>\n");
			tabs.append("</tr>\n");
			tabs.append("</table>");
			if (position != BOTTOM) sb.append(tabs);
		}
	
		if (content.length() > 0)
		{
			sb.append("<div id=\"").append(getId()).append(".content\"");
 			sb.append(" class=\"tabcontents\"");
 			
			sb.append(">");
			sb.append(content);
			sb.append("</div>");
		}
		if (!noTabs && position == BOTTOM) sb.append(tabs);	
		sb.append("</div>\n");
		numTabs = 0;

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}
	
	/*
		From the Renderer interface
		---------------------------
	*/
	@Override
	public void render(TabAction action)
	{
		addTab(action);
	}
	// ---------------------------
	
}


