/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.io.ByteArrayInputStream;
import java.security.KeyStore;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.http.conn.ssl.SSLConnectionSocketFactory;

import net.sf.basedb.core.data.FileServerData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.ssl.SSLUtil2;

/**
	This class represents a file server that is used to store
	external files. A {@link File} should be associated with a file
	server if authentication is required to access the file.
	
	@author Nicklas
	@since 2.16
	@base.modified $Date $
*/
public class FileServer 
	extends CommonItem
{
	
	/**
		The type of item represented by this class.
		@see Item#FILESERVER
		@see #getType()
	*/
	public static final Item TYPE = Item.FILESERVER;

	/**
		Pattern that matches a SSH fingerprint: 16 pairs of hexadecimal numbers separated with colon.
		@since 3.3
	*/
	public static final Pattern FINGERPRINT_PATTERN = Pattern.compile("[0-9a-f]{2}(\\:[0-9a-f]{2}){15}");
	
	/**
		Create a new <code>FileServer</code> item.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>FileServer</code> item
		@throws BaseException If there is an error
	*/
	public static FileServer getNew(DbControl dc)
		throws BaseException
	{
		FileServer fs = dc.newItem(FileServer.class);
		fs.setName("New server");
		return fs;
	}
	
	/**
		Get a <code>FileServer</code> item when you know the ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>FileServer</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static FileServer getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		FileServer fs = dc.loadItem(FileServer.class, id);
		if (fs == null) throw new ItemNotFoundException("FileServer[id="+id+"]");
		return fs;
	}
	
	/**
		Get a new query object for this class.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<FileServer> getQuery()
	{
		return new ItemQuery<FileServer>(FileServer.class);
	}

	FileServer(FileServerData data)
	{
		super(data);
	}

	@Override
	FileServerData getData()
	{
		return (FileServerData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Checks if there are files registered with this server.
		@return TRUE if this item is used, FALSE otherwise
		@throws BaseException If there is another error
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_FILES_FOR_FILESERVER", "count(*)");
		/*
			SELECT {1}
			FROM net.sf.basedb.core.data.FileData f
			WHERE f.fileServer = :fileServer
		*/
		query.setEntity("fileServer", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		return used || super.isUsed();
	}
	/**
		Get all files registered with this server.
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_FILES_FOR_FILESERVER", "f.id");
		/*
			SELECT {1}
			FROM net.sf.basedb.core.data.FileData f
			WHERE f.fileServer = :fileServer
		*/
		query.setEntity("fileServer", this.getData());
		addUsingItems(using, Item.FILE, query);
		return using;
	}
	// ----------------------------------------
	
	/**
		The maximum length of the connection manager factory id that can be stored in
		the database.
		@see #setConnectionManagerFactoryId(String)
	*/
	public static final int MAX_CONNECTION_MANAGER_FACTORY_LENGTH = FileServerData.MAX_CONNECTION_MANAGER_FACTORY_LENGTH;
	
	/**
		Get the ID of the connection manager factory that should be
		used to access file contents and metadata. A null value mean
		that auto-detection based on the file URI should be used.
		@return An ID or null if using auto-detection
	*/
	public String getConnectionManagerFactoryId()
	{
		return getData().getConnectionManagerFactory();
	}
	
	/**
		Set the ID of the connection manager factory.
		  
		@param factoryId The new ID of the factory, null if using auto-detection
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the ID is longer than {@link #MAX_CONNECTION_MANAGER_FACTORY_LENGTH}
	*/
	public void setConnectionManagerFactoryId(String factoryId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setConnectionManagerFactory(
				StringUtil.setNullableString(factoryId, "factoryId", MAX_CONNECTION_MANAGER_FACTORY_LENGTH));
	}
	
	/**
		The maximum length of the host name:port that can be stored in
		the database.
		@see #setHost(String)
		@since 3.0
	*/
	public static final int MAX_HOST_LENGTH = FileServerData.MAX_HOST_LENGTH;
	
	/**
		Get the name:port of the host that should override the host:port
		in the file url for files using this file server.

		@return A string or null if no override should be done.
		@since 3.0
	*/
	public String getHost()
	{
		return getData().getHost();
	}
	
	/**
		Set the host:port that should overrride the host:port
		specified in file urls.
		  
		@param host The host:port or null to not override
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the host is longer than {@link #MAX_HOST_LENGTH} or
			is an invalid host name + port
		@since 3.0
	*/
	public void setHost(String host)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (host != null) 
		{
			if (!host.matches("[\\p{L}\\d_-]+(\\.[\\p{L}\\d_-]+)*(\\:\\d+)?"))
			{
				throw new InvalidDataException("Not a valid host name: " + host);
			}
		}
		getData().setHost(StringUtil.setNullableString(host, "host", MAX_HOST_LENGTH));
	}
	
	/**
		The maximum length of the username that can be stored in the database.
		@see #setUsername(String)
	*/
	public static final int MAX_USERNAME_LENGTH = FileServerData.MAX_USERNAME_LENGTH;
	
	/**
		Get the username that is required to authenticate with the file
		server.  
		@return A username or null if authentication is not required
	*/
	public String getUsername()
	{
		return getData().getUsername();
	}
	
	/**
		Set the username that is required to authenticate with the file
		server.  
		@param username The new username, null if not authentication is required
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the username is longer than {@link #MAX_USERNAME_LENGTH}
	*/
	public void setUsername(String username)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setUsername(StringUtil.setNullableString(username, "username", MAX_USERNAME_LENGTH));
	}


	/**
		The maximum length of the password that can be stored in the database.
		@see #setPassword(String)
	*/
	public static final int MAX_PASSWORD_LENGTH = FileServerData.MAX_PASSWORD_LENGTH;

	/**
		The maximum length of the root path that can be stored in the database.
		@see #setRootPath(String)
		@since 3.3
	*/
	public static final int MAX_ROOT_PATH_LENGTH = FileServerData.MAX_ROOT_PATH_LENGTH;

	/**
		The length of the SSH fingerprint that can be stored in the database.
		@see #setSshFingerprint(String)
		@since 3.3
	*/
	public static final int FINGERPRINT_LENGTH = FileServerData.FINGERPRINT_LENGTH;

	
	/**
		Get the password that is used when authenticating with the file
		server. NOTE! Write permission is required to read the password.
		@return The password (which may be null)
		@see #hasPassword()
	*/
	public String getPassword()
	{
		if (!hasPermission(Permission.WRITE))
		{
			throw new PermissionDeniedException("WRITE permission is required to read the password");
		}
		return getData().getPassword();
	}
	
	/**
		Is a password set or not?
	*/
	public boolean hasPassword()
	{
		return getData().getPassword() != null;
	}
	
	/**
		Set the password to use when authenticating with the file server.
		
		@param password The new password (can be null)
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the password is longer than {@link #MAX_PASSWORD_LENGTH}
	*/
	public void setPassword(String password)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setPassword(StringUtil.setNullableString(password, "password", MAX_PASSWORD_LENGTH));
	}
	
	/**
		The maximum length of the server certificate that can be stored in the database.
		@see #setServerCertificate(byte[])
		@see #setClientCertificate(byte[])
	*/
	public static final int MAX_CERTIFICATE_LENGTH = FileServerData.MAX_CERTIFICATE_LENGTH;

	/**
		Get the public server certificate that indicates that we should only trust the 
		server if it can present a certificate that matches this. This property is typically
		only needed to access servers using https that has a self-signed certificate. Servers
		which has a valid certificate signed by a trusted certification authority should be
		automatically trusted by BASE without setting this property. The certificate
		must be a X.509 certificate in either binary or base64-encoded DER format.
		
		@return A byte array with the certificate or null if no certificate has been set
		@see #getServerX509Certificate()
	*/
	public byte[] getServerCertificate()
	{
		byte[] cert = getData().getServerCertificate();
		if (cert == null || cert.length == 0) return null;
		return Arrays.copyOf(cert, cert.length);
	}
	
	/**
		Set the public server certificate so that BASE can trust https connections to
		the server. See {@link #getServerCertificate()} for more information. The certificate
		must be a X.509 certificate in either binary or base64-encoded DER format.
		@param certificate The server certificate or null if no certificate is required
	*/
	public void setServerCertificate(byte[] certificate)
	{
		checkPermission(Permission.WRITE);
		if (certificate != null && certificate.length > MAX_CERTIFICATE_LENGTH)
		{
			throw new InvalidDataException("The 'serverCertificate' mustn't be longer than "+MAX_CERTIFICATE_LENGTH+
					" bytes. The supplied value is "+certificate.length+ " bytes.");
		}
		byte[] cert = null;
		if (certificate != null && certificate.length > 0)
		{
			cert = Arrays.copyOf(certificate, certificate.length);
		}
		getData().setServerCertificate(cert);
	}
	
	/**
		Get the X.509 certificate for the server.
		@return The server certificate or null if no certificate has been specified
	*/
	public X509Certificate getServerX509Certificate() 
	{
		byte[] cert = getData().getServerCertificate();
		if (cert == null || cert.length == 0) return null;
		try
		{
			return (X509Certificate)SSLUtil2.getCertificate(new ByteArrayInputStream(cert), "X.509", null);
		}
		catch (Exception ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Get the client certificate that BASE should use to authenticate with the server
		when connecting with https. This property is typically needed when the server
		requires clients to authenticate themselves using a certificate. The certificate
		must be a PKCS #12 certificate in binary format, and it should contain only
		one certificate. NOTE! WRITE permission is required to read this property since
		it may contain sensitive data, use {@link #getClientX509Certificate()} to get 
		the public information from the certificate.
		
		@return A byte array with the certificate or null if no certificate has been set
		@see #getClientX509Certificate()
	*/
	public byte[] getClientCertificate()
	{
		if (!hasPermission(Permission.WRITE))
		{
			throw new PermissionDeniedException("WRITE permission is required to read the raw client certificate");
		}
		byte[] cert = getData().getClientCertificate();
		if (cert == null || cert.length == 0) return null;
		return Arrays.copyOf(cert, cert.length);
	}
	
	/**
		Set the client certificate that BASE should use to authenticate with the server
		when connecting with https. The certificate must be a PKCS #12 certificate in binary 
		format, and it should contain only one certificate.
		
		@param certificate The client certificate or null if no certificate is required
	*/
	public void setClientCertificate(byte[] certificate)
	{
		checkPermission(Permission.WRITE);
		if (certificate != null && certificate.length > MAX_CERTIFICATE_LENGTH)
		{
			throw new InvalidDataException("The 'clientCertificate' mustn't be longer than "+MAX_CERTIFICATE_LENGTH+
					" bytes. The supplied value is "+certificate.length+ " bytes.");
		}
		byte[] cert = null;
		if (certificate != null && certificate.length > 0)
		{
			cert = Arrays.copyOf(certificate, certificate.length);
		}
		getData().setClientCertificate(cert);
	}
	
	/**
		Get the X.509 certificate for the client.
		@return The client certificate or null if no certificate has been specified
	*/
	public X509Certificate getClientX509Certificate() 
	{
		byte[] key = getData().getClientCertificate();
		if (key == null || key.length == 0) return null;
		
		try
		{
			KeyStore ks = SSLUtil2.createKeyStore(new ByteArrayInputStream(key), getClientCertificatePassword(), "PKCS12", null);
			return (X509Certificate)ks.getCertificate(ks.aliases().nextElement());
		}
		catch (Exception ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Get the password that is needed to unlock the client
		certificate. NOTE! WRITE permission is required to read this property.
		@return The password (which may be null)
	*/
	public String getClientCertificatePassword()
	{
		if (!hasPermission(Permission.WRITE))
		{
			throw new PermissionDeniedException("WRITE permission is required to read the password");
		}
		return getData().getClientCertificatePassword();
	}
	
	/**
		Set the password that is needed to unlock the client
		certificate.
		
		@param password The new password (can be null)
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the password is longer than {@link #MAX_PASSWORD_LENGTH}
	*/
	public void setClientCertificatePassword(String password)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setClientCertificatePassword(StringUtil.setNullableString(password, 
				"clientCertificatePassword", MAX_PASSWORD_LENGTH));
	}
	
	/**
		@since 3.3
		@see #setRootPath(String)
	*/
	public String getRootPath()
	{
		return getData().getRootPath();
	}
	
	/**
		Set the root path that should be added to all URI:s when used with this file
		server. Eg. if URI specify http://myhost.com/path/to/file it is converted
		to http://myhost.com/rootPath/path/to/file
		@param rootPath The root path or null to not use any root path
		@since 3.3
	*/
	public void setRootPath(String rootPath)
	{
		checkPermission(Permission.WRITE);
		if (rootPath != null)
		{
			if (!rootPath.startsWith("/"))
			{
				throw new InvalidDataException("rootPath must start with '/'");
			}
			if (rootPath.endsWith("/"))
			{
				throw new InvalidDataException("rootPath must not end with '/'");
			}
		}
		getData().setRootPath(StringUtil.setNullableString(rootPath, 
				"rootPath", MAX_ROOT_PATH_LENGTH));
	}
	
	/**
		@since 3.3
		@see #setSshFingerprint(String)
	*/
	public String getSshFingerprint()
	{
		return getData().getSshFingerprint();
	}
	
	/**
		Set the SSH fingerprint for the server. Connections made to the server
		with SSH should be verified against this fingerprint to protect against
		man-in-the-middle attacks. If no fingerprint is present, the implementation
		may choose to reject or accept all connections.
		@param fingerprint The fingerprint or null to not use fingerprint
		@since 3.3
	*/
	public void setSshFingerprint(String fingerprint)
	{
		checkPermission(Permission.WRITE);
		if (fingerprint != null)
		{
			if (fingerprint.length() != FINGERPRINT_LENGTH)
			{
				throw new InvalidDataException("SSH fingerprint must be exactly " + FINGERPRINT_LENGTH + " characters long");
			}
			if (!FINGERPRINT_PATTERN.matcher(fingerprint).matches())
			{
				throw new InvalidDataException("Not a valid SSH fingerprint: " + fingerprint);
			}
		}
		getData().setSshFingerprint(fingerprint);
	}

	/**
		Get a socket factory that can be used to create SSL connections to the given
		server. If a server and/or client certificate has been specified for this
		file server the socket factory will be created with those certificates.
		Otherwise, the default (as configured in base.config) certificates are used.
		@return A SSL socket factory
	*/
	public SSLConnectionSocketFactory getSSLSocketFactory2()
	{
		return SSLUtil2.getSSLSocketFactory(getServerCertificate(), getClientCertificate(), getClientCertificatePassword());
	}
	
	/**
		Get a query returning all files that has been registered with this
		server.
	*/
	public ItemQuery<File> getFiles()
	{
		ItemQuery<File> query = File.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("fileServer"), Hql.entity(this))
			);
		return query;
	}
	
}
