/*
	$Id: MeasuredBioMaterial.java 7003 2015-11-06 13:08:14Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import net.sf.basedb.core.data.BioMaterialEventData;
import net.sf.basedb.core.data.BioPlateData;
import net.sf.basedb.core.data.BioPlateTypeData;
import net.sf.basedb.core.data.BioWellData;
import net.sf.basedb.core.data.MeasuredBioMaterialData;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.LoggingInterceptor;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.collections.BasicItemToIdTransformer;

/**
	This class is the base class for measured biomaterial items. A measured
	item is an item which can have original and remaining quantities
	associated with them.
	<p>
	The core keeps track of the remaining quantity whenever an item is used to create other items.
	As it is, an item can be used when creating another item of the same type
	(pooling) or when creating an item of the appropriate subtype, ie.
	<code>Sample -&gt; Extract -&gt; LabeledExtract</code>.
	<p>
	It is also possible to add a note using a {@link BioMaterialEvent} 
	saying that the quantity has been changed and why.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-11-06 14:08:14 +0100 (fr, 06 nov 2015) $
*/
public abstract class MeasuredBioMaterial
	extends BioMaterial
	implements Registered
{

	private BioPlateEventParticipant bioPlateEventParticipant;
	private BioMaterialEvent creationEvent;

	MeasuredBioMaterial(MeasuredBioMaterialData measuredBioMaterialData)
	{
		super(measuredBioMaterialData);
		if (measuredBioMaterialData.getId() == 0) measuredBioMaterialData.setRemainingQuantity(0.0f);
	}
	
	@Override
	MeasuredBioMaterialData getData()
	{
		return (MeasuredBioMaterialData)getBasicData();
	}

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>no pooled biomaterial has been created from this item
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SOURCEEVENTS_FOR_BIOMATERIAL", "count(*)");
		/*
			SELECT {1}
			FROM BioMaterialEventData bme
			JOIN bme.sources src
			WHERE src.bioMaterial = :bioMaterial
		*/
		query.setEntity("bioMaterial", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link BioMaterial}:s where this item has been used as a source
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SOURCEEVENTS_FOR_BIOMATERIAL", "bme.bioMaterial");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				JOIN bme.sources src
				WHERE src.bioMaterial = :bioMaterial
			*/
		query.setEntity("bioMaterial", this.getData());
		addUsingItems(using, query);
		return using;
	}

	/**
		Make sure the creation event is created at the same time this item
		is created and that source biomaterials get their used quantities back
		if this item is deleted. All events are deleted by cascade mapping in Hibernate.
	*/
	@Override
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.CREATE)
		{
			getDbControl().saveItem(getCreationEvent());
		}
		else if (action == Transactional.Action.DELETE)
		{
			if (getData().getBioWell() != null)
			{
				getData().getBioWell().getBioPlate().updateFreeWells(1);
			}
			getCreationEvent().clearSources();
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(getDbControl().getHibernateSession(), 
					"CLEAR_ORIGINAL_BIOMATERIAL_ON_BIOWELLS");
			/*
				UPDATE BioWellData bw
				SET bw.originalBioMaterial = null
				WHERE bw.originalBioMaterial = :bioMaterial
			*/
			query.setEntity("bioMaterial", getData());
			query.executeUpdate();
		}
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		Get the protocol from the creation event.
		@since 2.2
	*/
	@Override
	public Protocol getProtocol()
	{
		return getCreationEvent().getProtocol();
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return getCreationEvent().getEntryDate();
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/** 
		Get the biosource or pooled samples and the bioplate 
		the biomaterial is located on.
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		return getAnnotatableParents(null);
	}
	// -------------------------------------------

	/**
		Get the annotable parents, possible only a given set of specific parents items.
		@param parents If not null or empty, only include the given parent items (for pooled items only)
		@since 3.3
	*/
	public Set<Annotatable> getAnnotatableParents(Collection<? extends MeasuredBioMaterial> parents)
		throws BaseException
	{
		Set<Annotatable> annotatable = super.getAnnotatableParents();
		DbControl dc = getDbControl();
		
		MeasuredBioMaterialData data = getData();
		BioMaterialEvent event = getCreationEvent();
		BioMaterialEventData eventData = event.getData();

		addAnnotatableParents(dc, annotatable, 
			eventData.getHardware(),
			eventData.getProtocol(),
			eventData.getKit()
		);
		
		if (data.getBioWell() != null)
		{
			addAnnotatableParents(dc, annotatable, data.getBioWell().getBioPlate());
		}

		if (data.getParent() != null)
		{
			// Single parent item
			addAnnotatableParents(dc, annotatable, data.getParent());
		}
		else if (getParentType() != null)
		{
			// Multiple parent items
			ItemQuery<? extends BioMaterial> query = event.getSources();
			query.include(Include.ALL);
			if (parents != null && parents.size() > 0)
			{
				query.restrict(Restrictions.in(Hql.property("id"), Expressions.parameter("parents")));
				Set<Integer> parentIds = new HashSet<Integer>();
				new BasicItemToIdTransformer().transform(parents, parentIds);
				query.setParameter("parents", parentIds, Type.INT);
			}
			annotatable.addAll(query.list(dc));
		}
		
		return annotatable;
	}

	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	@SubtypableRelatedItems({Item.PROTOCOL, Item.HARDWARE})
	public ItemSubtype getItemSubtype()
	{
		return super.getItemSubtype();
	}
	// -------------------------------------------
	
	/**
		Get the event that represents the creation of this biomaterial.
		@return A {@link BioMaterialEvent} data.
	 	@throws PermissionDeniedException If logged in user doesn't have read permission on the biomaterial. 
	 	@throws BaseException If there is some other error. 
	*/
	public BioMaterialEvent getCreationEvent()
		throws PermissionDeniedException, BaseException
	{
		if (creationEvent == null)
		{
			DbControl dc = getDbControl();
			if (!isInDatabase())
			{
				creationEvent = BioMaterialEvent.getNewCreationEvent(dc, this, bioPlateEventParticipant);
			}
			else
			{
				creationEvent = getDbControl().getItem(BioMaterialEvent.class, getData().getCreationEvent());
			}
		}
		return creationEvent;
	}
	
	void setBioPlateEventParticipant(BioPlateEventParticipant bioPlateEventParticipant)
	{
		this.bioPlateEventParticipant = bioPlateEventParticipant;
	}
	
	/**
		Get the original quantity of the biomaterial.
		@return A <code>Float</code> holding the original quantity, or null
			if not known
	*/
	public Float getOriginalQuantity()
	{
		return getData().getOriginalQuantity();
	}

	/**
		Set the original quantity of the biomaterial. If the remaining quantity
		is null, it's value is also set to the same as the original value.
		@param originalQuantity The new value for the original quantity, or
			null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the original quantity is below zero
		@throws BaseException If there is some other kind of error.
	*/
	public void setOriginalQuantity(Float originalQuantity)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (originalQuantity != null && originalQuantity < 0.0f) 
		{
			throw new NumberOutOfRangeException("originalQuantity", originalQuantity, 0.0f, false);
		}
		getData().setOriginalQuantity(originalQuantity);
		getCreationEvent().setUsedQuantity(originalQuantity == null ? null : -originalQuantity);
	}

	/**
		Get the remaining quantity of the biomaterial. The remaining quantity 
		cannot be modified directly. It is calculated based on the events for this 
		biomaterial. Use the {@link #newEvent()} method to get an event object where you
		can specify the used quantity.
		@return A <code>Float</code> holding the remaining quantity or null
			if not known
	*/
	public Float getRemainingQuantity()
	{
		MeasuredBioMaterialData data = getData();
		return data.getOriginalQuantity() == null ? null : data.getRemainingQuantity();
	}
	
	/**
		Update the remaining quantity of a biomaterial. This method does nothing
		if the <code>usedQuantity</code> parameter is null or if the remaining
		quantity is null.
	*/
	static void updateRemainingQuantity(DbControl dc, MeasuredBioMaterialData bioMaterial, Float usedQuantity)
	{
		if (usedQuantity == null) return;
		if (bioMaterial.getId() == 0)
		{
			// For unsaved, objects we must use the object itself
			Float remain = bioMaterial.getRemainingQuantity();
			if (remain == null) remain = 0.0f;
			bioMaterial.setRemainingQuantity(remain-usedQuantity);
		}
		else
		{
			// For objects in the database we use an UPDATE query to solve transaction problems
			RemainingQuantityAction action = RemainingQuantityAction.get(dc, bioMaterial);
			action.addUsedQuantity(usedQuantity);
		}
	}
	
	/**
		Get the item type of the parent biomaterial. The parent type
		must be either the same type as the current biomaterial or "one
		step up" in the Extract -- Sample -- BioSource chain.
		@return The parent type, or null if the biomaterial doesn't have any parents
		@since 3.0
	*/
	public Item getParentType()
	{
		Integer parentType = getData().getParentType();
		return parentType == null ? null : Item.fromValue(parentType);
	}

	/**
		Get the parent biomaterial if there is a single parent item only.
		@return The parent biomaterial or null if there is not exactly one
			parent
		@since 3.0
		@see #hasSingleParent()
	 */
	public BioMaterial getParent()
	{
		return getDbControl().getItem(BioMaterial.class, getData().getParent());
	}
	
	/**
		Checks if this biomaterial has a single parent item or not.
		@return TRUE if there is exactly one parent, FALSE if there are
			0 or 2 or more parents
		@since 3.0
		@see #getParentType()
	*/
	public boolean hasSingleParent()
	{
		return getData().getParent() != null;
	}
	
	/**
		Create a new event for this biomaterial.
	 	@return a {@link net.sf.basedb.core.BioMaterialEvent} object. 
		@throws PermissionDeniedException If the logged in user doesn't have 
			use permission for this biomaterial
		@throws BaseException If there is another error
	*/
	public BioMaterialEvent newEvent()
		throws PermissionDeniedException, BaseException
	{
		return BioMaterialEvent.getNew(getDbControl(), this);
	}

	/**
		Create a new event for this biomaterial that is part of a
		master bioplate event.
	 	@return a {@link net.sf.basedb.core.BioMaterialEvent} object. 
		@throws PermissionDeniedException If the logged in user doesn't have 
			use permission for this biomaterial
		@throws BaseException If there is another error
		@since 2.17
	*/
	public BioMaterialEvent newEvent(BioPlateEventParticipant master)
		throws PermissionDeniedException, BaseException
	{
		return BioMaterialEvent.getNew(getDbControl(), this, master);
	}
	
	/**
		Get a query that returns all events for this biomaterial.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<BioMaterialEvent> getEvents()
	{
		return BioMaterialEvent.getQuery(this);
	}
	
	/**
	 	Sets the <code>BioWell</code> where this biomaterial is located. 
	 	
	  	@param well The <code>BioWell</code> to set.
	  	@throws ItemInUseException If the well is already used by another
	  		biomaterial
	  	@throws PermissionDeniedException If logged in user doesn't have 
	  		{@link Permission#WRITE} for this MeasuredBioMaterial or
	  		{@link Permission#USE} for the BioWell.
	  	@throws BaseException If there is another error.
	  	@since 2.10
	*/
	public void setBioWell(BioWell well)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		MeasuredBioMaterialData myData = getData();
		BioWellData currentWell = myData.getBioWell();
		BioWellData newWell = well != null ? well.getData() : null;
		if (EqualsHelper.equals(currentWell, newWell)) return; // early exit if no change
		
		// Check if current biowell is locked
		if (currentWell != null && currentWell.getId() != 0)
		{
			BioPlateData currentPlate = currentWell.getBioPlate();
			BioWell.LockMode lockMode = BioWell.LockMode.fromValue(currentPlate.getBioPlateType().getLockMode());
			
			if (!lockMode.canClear(currentWell))
			{
				throw new PermissionDeniedException(Permission.WRITE, this + 
						". Well on plate '" + currentPlate.getName() + "'' is locked: " + currentWell);
			}
			lockMode.onClear(currentWell, myData);
		}
		
		if (well != null)
		{
			// Check that the current user may use the well
			well.checkPermission(Permission.USE);
			BioPlateData newPlate = newWell.getBioPlate();
			// Check that the new well is empty
			if (newWell.getBioMaterial() != null)
			{
				throw new ItemInUseException("Well [" + well.getRow() + "," + well.getColumn() + 
					"] on plate '" + newPlate.getName() + "'");
			}
			// Check that the plate type allows the current type of biomaterial
			BioPlateTypeData newPlateType = newPlate.getBioPlateType();
			Integer biomaterialType = newPlateType.getBioMaterialType();
			if (biomaterialType != null && biomaterialType != getType().getValue())
			{
				throw new PermissionDeniedException(Permission.USE, 
					"Plate '" + newPlate.getName() + "' can only be used with " +
					Item.fromValue(biomaterialType));
			}
			// Check that the lock mode allows adding biomaterial to the well
			BioWell.LockMode lockMode = BioWell.LockMode.fromValue(newPlateType.getLockMode());
			if (!lockMode.canAdd(newWell))
			{
				throw new PermissionDeniedException(Permission.WRITE, this + 
						". Well on plate '" + newPlate.getName() + "'' is locked: " + well);
			}
			lockMode.onAdd(newWell, myData);
		}
		if (currentWell != null) 
		{
			currentWell.setBioMaterial(null);
			currentWell.getBioPlate().updateFreeWells(1);
		}
		if (newWell != null) 
		{
			newWell.setBioMaterial(myData);
			newWell.getBioPlate().updateFreeWells(-1);
		}
		myData.setBioWell(newWell);
	}
	
	/**
		Get the well this biomaterial is located in.
		@return A BioWell object or null if this biomaterial is not located on
			a plate
		@since 2.10
	*/
	public BioWell getBioWell()
	{
		return getDbControl().getItem(BioWell.class, getData().getBioWell());
	}
	
	/**
		Check if this biomaterial is located in a well on a bioplate or not.
		@return TRUE if the biomaterial is in a well, FALSE if not
		@since 2.17
	*/
	public boolean isInWell()
	{
		return getData().getBioWell() != null;
	}
	
	/**
		Check if this biomaterial is placed in a well that it can't be removed from.
		
		@return TRUE if the biomaterial is locked, FALSE if is not in a well or in
			a well that allows moving it
		@since 2.16
	*/
	public boolean isLockedInWell()
	{
		MeasuredBioMaterialData myData = getData();
		BioWellData currentWell = myData.getBioWell();
		
		// Not locked if not in a well
		if (currentWell == null) return false;

		// Get the lock mode from the plate's plate type
		BioPlateData plate = currentWell.getBioPlate();
		BioWell.LockMode lockMode = BioWell.LockMode.fromValue(plate.getBioPlateType().getLockMode());
		
		// The lock mode knows if we can move the biomaterial
		return !lockMode.canClear(currentWell);
	}
	
	/**
		Transaction handler for updating the remaining quantity of a biomaterial
		based on the sum of the used quantity in this transaction.
		@since 3.3
	*/
	static class RemainingQuantityAction
		implements TransactionalAction
	{

		/**
			We need a single action per DbControl and BioMaterial
		*/
		private static Map<String, RemainingQuantityAction> cache = 
			Collections.synchronizedMap(new WeakHashMap<String, RemainingQuantityAction>());
		
		private final DbControl dc;
		private final MeasuredBioMaterialData bioMaterial;
		private float usedQuantity;
		
		static RemainingQuantityAction get(DbControl dc, MeasuredBioMaterialData bioMaterial)
		{
			String key = dc.hashCode() + ":" + bioMaterial.getId();
			RemainingQuantityAction action = cache.get(key);
			if (action == null)
			{
				action = new RemainingQuantityAction(dc, bioMaterial);
				dc.addTransactionalAction(action);
				cache.put(key, action);
			}
			return action;
		}
		
		private RemainingQuantityAction(DbControl dc, MeasuredBioMaterialData bioMaterial)
		{
			this.dc = dc;
			this.bioMaterial = bioMaterial;
		}
		
		/*
			From the TransactionalAction interface
		    --------------------------------------
		*/
		@Override
		public void onBeforeCommit() 
		{
			// Load current value from the database
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "GET_REMAINING_QUANTITY");
			query.setEntity("bioMaterial", bioMaterial);
			Float remainBefore = HibernateUtil.loadData(Float.class, query);
			
			// Make the change in the database
			query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "UPDATE_REMAINING_QUANTITY");
			/*
				UPDATE MeasuredBioMaterialData mbm 
				SET mbm.remainingQuantity = mbm.remainingQuantity - :used
				WHERE mbm = :bioMaterial
			*/
			query.setFloat("used", usedQuantity);
			query.setEntity("bioMaterial", bioMaterial);
			HibernateUtil.executeUpdate(query);

			// Notify the logging system
			LoggingInterceptor logInterceptor = dc.getLoggingInterceptor();
			if (logInterceptor != null)
			{
				Float remainAfter = remainBefore - usedQuantity;
				logInterceptor.addManualSingleChange(ChangeType.UPDATE, bioMaterial, remainAfter, 
					remainBefore, "remainingQuantity", Type.FLOAT.getTypeWrapper().getHibernateType());
			}
			removeFromCache();
		}

		@Override
		public void onAfterCommit() 
		{
			removeFromCache();
		}

		@Override
		public void onRollback() 
		{
			removeFromCache();
		}
		// --------------------------------------
		
		private void removeFromCache()
		{
			String key = dc.hashCode() + ":" + bioMaterial.getId();
			cache.remove(key);
		}
		
		/**
			Add some used quantity to what has currently been recorded.
		*/
		void addUsedQuantity(float usedQuantity)
		{
			this.usedQuantity += usedQuantity;
		}
		
	}
}