/*
	$Id: Permission.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.core;

import java.util.EnumSet;
import java.util.Set;
/**
	This enumeration defined constants for permissions.
	
	@base.developer
	Permissions are stored in the database as integers, which means that
	a given set of permissions must be combined before they are saved.
	We use the <code>grantValue</code> for this. This value is constructed
	in such a way that some permissions implicitly activates other permissions
	as follows:
	<ul>
	<li>READ -&gt; 
	<li>USE -&gt; READ
	<li>RESTRICTED_WRITE -&gt; USE and READ
	<li>WRITE -&gt; RESTRICTED_WRITE, USE and READ
	<li>DELETE -&gt; WRITE, RESTRICTED_WRITE, USE and READ
	<li>SET_OWNER -&gt; WRITE, RESTRICTED_WRITE, USE and READ
	<li>SET_PERMISSION -&gt; WRITE, RESTRICTED_WRITE, USE and READ
	</ul>
	
	The <code>denyValue</code> is only used by the {@link BasicItem#initPermissions(int, int)}
	method when a subclass needs to deny a permission. This value is constructed 
	in a similiar way:
	<ul>
	<li>READ -&gt; USE, RESTRICTED_WRITE, WRITE, DELETE, SET_OWNER and SET_PERMISSION
	<li>USE -&gt; RESTRICTED_WRITE, WRITE, DELETE, SET_OWNER and SET_PERMISSION
	<li>RESTRICTED_WRITE -&gt; WRITE, DELETE, SET_OWNER and SET_PERMISSION
	<li>WRITE -&gt; DELETE, SET_OWNER and SET_PERMISSION
	<li>DELETE -&gt; 
	<li>SET_OWNER -&gt; 
	<li>SET_PERMISSION -&gt; 
	</ul>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-21 09:59:42 +0200 (ti, 21 apr 2015) $
*/
public enum Permission
{
	/**
		This permission allows a user read information about an item.
	*/
	READ(1, 1+2+4+8+16+32, "read"),
	
	/**
		This permission allows a user to use/link to an item.
	*/
	USE(1+2, 2+4+8+16+32, "use"),
	
	/**
		This permission allows a user to update some information about an item.
	*/
	RESTRICTED_WRITE(1+2+4, 4+8+16+32+64, "restricted write"),

	/**
		This permission allows a user update the information about an item.
	*/
	WRITE(1+2+4+8, 8+16+32+64, "write"),
	
	/**
		This permission allows a user delete an item.
	*/
	DELETE(1+2+4+8+16, 16, "delete"),
	
	/**
		This permission allows a user change the owner of an item.
	*/
	SET_OWNER(1+2+4+8+32, 32, "change owner"),
	
	/**
		This permission allows a user change the access permission to an item.
	*/
	SET_PERMISSION(1+2+4+8+64, 64, "set permissions"),
	
	/**
		This permission allows a user create new items.
	*/
	CREATE(128, 128, "create"),
	
	/**
		This permission denies a user access to an item. This permission can only be given
		to {@link RoleKey}:s.
	*/
	DENIED(256, 256, "denied"),
	
	/**
		This permission allows a user to share an item to the {@link Group#EVERYONE}
		group. This is a system permission and is only meaningful for the role key
		for the {@link Item#SYSTEM} item.
	*/
	SHARE_TO_EVERYONE(512, 512, "share to everyone"),

	/**
		This permission allows a user act as another user using the {@link
		SessionControl#impersonateLogin(int, String)} method. This is a system 
		permission and is only meaningful for the role key for the {@link Item#SYSTEM} 
		item.
	*/
	ACT_AS_ANOTHER_USER(1024, 1024, "act as another user"),

	/**
		This permission allows a user to specify which job agent
		a job should be executed on using {@link Job#setJobAgent(JobAgent)}.
		This is a system permission and is only meaningful for the role key 
		for the {@link Item#SYSTEM} item.
		@since 2.8
	*/
	SELECT_JOBAGENT(2048, 2048, "select job agent");

	private final int grantValue;
	private final int denyValue;
	private final String displayValue;
	
	/**
		Constructor for Permission enums.
	*/
	private Permission(int grantValue, int denyValue, String displayValue)
	{
		this.grantValue = grantValue;
		this.denyValue = denyValue;
		this.displayValue = displayValue;
	}
	int grantValue()
	{
		return grantValue;
	}
	int denyValue()
	{
		return denyValue;
	}

	@Override
	public String toString()
	{
		return displayValue;
	}

	/**
		Expand the given set of permissions to make sure it includes
		all implicitely granted permissions. For example if the set contains 
		WRITE it is exapnded to READ, USE, RESTRICTED_WRITE and WRITE.
		
		@param permissions The permissions to expand
		@return A new set containing the expanded permissions or null 
			if the parameter is null
	*/
	public static Set<Permission> expand(Set<Permission> permissions)
	{
		if (permissions == null) return null;
		return fromInt(grant(permissions));
	}
	
	/**
		Checks if the given <code>Permission</code> is granted by the 
		<code>permissions</code> code.
		@param permissions The permission combination
		@param permission The permission to check
	*/
	static boolean hasPermission(int permissions, Permission permission)
	{
		return (permissions & permission.grantValue()) == permission.grantValue();
	}
	
	/**
		Checks if the given <code>Permission</code> is denied by the 
		<code>permissions</code> code.
		@param permissions The permission combination
		@param permission The permission to check
	*/
	static boolean isDeniedPermission(int permissions, Permission permission)
	{
		return (permissions & permission.denyValue()) == permission.denyValue();
	}
	
	/**
		Convert the given permission to the integer grant value.
	*/
	static int grant(Permission permission)
	{
		return permission.grantValue();
	}
	
	/**
		Combine the given permissions and convert to the integer grant value.
	*/
	static int grant(Permission... permissions)
	{
		int result = 0;
		for (Permission p : permissions)
		{
			result |= p.grantValue();
		}
		return result;
	}

	/**
		Combine the given permission and convert to the integer grant value.
	*/
	static int grant(Set<Permission> permissions)
	{
		int result = 0;
		if (permissions != null)
		{
			for (Permission p : permissions)
			{
				result |= p.grantValue();
			}
		}
		return result;
	}
	
	/**
		Convert the given permission to the integer deny value.
	*/
	static int deny(Permission permission)
	{
		return permission.denyValue();
	}

	/**
		Combine the given permissions and convert to the integer deny value.
	*/
	static int deny(Permission... permissions)
	{
		int result = 0;
		for (Permission p : permissions)
		{
			result |= p.denyValue();
		}
		return result;
	}

	/**
		Combine the given permission and convert to the integer deny value.
	*/
	static int deny(Set<Permission> permissions)
	{
		int result = 0;
		if (permissions != null)
		{
			for (Permission p : permissions)
			{
				result |= p.denyValue();
			}
		}
		return result;
	}

	/**
		Convert an integer value to a set of granted permissions.
	*/
	static Set<Permission> fromInt(int permissions)
	{
		EnumSet<Permission> result = EnumSet.noneOf(Permission.class);
		for (Permission p : Permission.values())
		{
			if (hasPermission(permissions, p))
			{
				result.add(p);
			}
		}
		return result;
	}
	
	/**
		Convert an integer value to a set of denied permissions.
	*/
	static Set<Permission> fromDeniedInt(int permissions)
	{
		EnumSet<Permission> result = EnumSet.noneOf(Permission.class);
		for (Permission p : Permission.values())
		{
			if (isDeniedPermission(permissions, p))
			{
				result.add(p);
			}
		}
		return result;
		
	}

	/**
		Merges a list of integer permission values.
		@param permissions The permission values, null values are ignored
		@return The combined permissions
		@since 2.8
	*/
	static int merge(Integer... permissions)
	{
		int perm = 0;
		for (Integer i : permissions)
		{
			if (i != null) perm |= i;
		}
		return perm;
	}
	
}

