/*
	$Id: Well.java 7014 2015-11-12 08:43:25Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.WellData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Set;
import java.util.HashSet;

/**
	A well is located on a {@link Plate} and contains a {@link ReporterData}.
	If the plate the well is located on was created from some other plates
	via a {@link PlateMapping} the well also contains a reference back
	to the parent well on one of the source plates.
	<p>
	Once the well has been saved to the database it cannot be modified,
	except for annotations.
	
	@author Nicklas
	@version 2.0
	@see Plate
	@base.modified $Date: 2015-11-12 09:43:25 +0100 (to, 12 nov 2015) $
*/
public class Well
	extends ChildItem
	implements Annotatable
{
	/**
		The type of item represented by this class.
		@see Item#WELL
		@see #getType()
	*/
	public static final Item TYPE = Item.WELL;
	
	/**
		Get a <code>Well</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Well</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Well getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Well w = dc.loadItem(Well.class, id);
		if (w == null) throw new ItemNotFoundException("Well[id="+id+"]");
		return w;
	}

	/**
		Get a query configured to retrieve wells for a plate.
		
		@param plate The plate to retreive wells for, null is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If required parameter is null. 
		@see Plate#getEvents()
	*/
	public static ItemQuery<Well> getQuery(Plate plate)
		throws InvalidDataException
	{
		if (plate == null) throw new InvalidUseOfNullException("plate");
		ItemQuery<Well> query = new ItemQuery<Well>(Well.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("plate"),
				Hql.entity(plate)
			)
		);
		return query;
	}

	Well(WellData wellData)
	{
		super(wellData);
	}

	@Override
	WellData getData()
	{
		return (WellData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	@Override
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		DbControl dc = getDbControl();
		AnnotationSet as = dc.getItem(AnnotationSet.class, getData().getAnnotationSet(), this);
		if (as == null)
		{
			as = AnnotationSet.getNew(dc, this);
			getData().setAnnotationSet(as.getData());
			dc.saveItemIf(this, as, false);
		}
		return as;
	}
	@Override
	public boolean isAnnotated()
	{
		return getData().getAnnotationSet() != null;
	}
	@Override
	public void removeAnnotations()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationSetData as = getData().getAnnotationSet();
		if (as != null)
		{
			HibernateUtil.deleteData(getDbControl().getHibernateSession(), as);
			getData().setAnnotationSet(null);
		}
	}
	/**
		Get the parent well, if any.
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		DbControl dc = getDbControl();
		WellData data = getData();
		
		addAnnotatableParents(dc, annotatable, 
			data.getParent(),
			data.getPlate()
		);
		return annotatable;
	}
	/**
		@return Always null
		@since 2.2
	*/
	@Override
	public Protocol getProtocol()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>a {@link Feature} is linked to this well
		<li>another {@link Well} has this well as it's parent well
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FEATURES_FOR_WELL", "count(*)");
		/*
			SELECT {1}
			FROM FeatureData f
			WHERE f.well = :well
		*/
		query.setEntity("well", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_CHILDWELLS_FOR_WELL", "count(*)");
			/*
				SELECT {1}
				FROM Well w
				WHERE w.parent = :well
			*/
			query.setEntity("well", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>Child wells
		<li>Features derived from this well
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Child wells
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_CHILDWELLS_FOR_WELL", "w.id");
			/*
				SELECT {1}
				FROM Well w
				WHERE w.parent = :well
			*/
		query.setEntity("well", this.getData());
		addUsingItems(using, Item.WELL, query);
		
		// Features
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FEATURES_FOR_WELL", "f.id");
			/*
				SELECT {1}
				FROM FeatureData f
				WHERE f.well = :well
			*/
		query.setEntity("well", this.getData());
		addUsingItems(using, Item.FEATURE, query);
			
		return using;
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getPlate();
	}
	// -------------------------------------------

	/**
		Get the {@link Plate} this event belongs to.
		@return The <code>Plate</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate
		@throws BaseException If there is another error
	*/
	public Plate getPlate()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Plate.class, getData().getPlate());
	}
	
	/**
		Get the row coordinate of this well on the plate.
	*/
	public int getRow()
	{
		return getData().getRow();
	}

	/**
		Get the column coordinate of this well on the plate.
	*/
	public int getColumn()
	{
		return getData().getColumn();
	}
	
	/**
		Get the parent well. The parent well is one of the wells on
		one of the parent plates if the plate this well is on was created
		by a {@link PlateMapping}.
		@return A <code>Well</code> item or null if this plate has no
			parents
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission to the parent well
		@throws BaseException If there is another error
	*/
	public Well getParent()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Well.class, getData().getParent());
	}
	
	/**
		Set the parent well. Used by {@link PlateMapping} when
		creating new plates from a mapping. This method also sets
		the reporter to the same reporter as the parent well.
	*/
	void setParent(Well parent)
	{
		getData().setParent(parent == null ? null : parent.getData());
		getData().setReporter(parent == null ? null : parent.getData().getReporter());
	}
	
	/**
		Get the reporter in this well.
		@return A <code>ReporterData</code> object or null if no 
			reporter has been specified
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission to reporters
		@throws BaseException If there is another error
	*/
	public ReporterData getReporter()
		throws PermissionDeniedException, BaseException
	{
		if (getSessionControl().hasPermission(Permission.DENIED, Item.REPORTER))
		{
			throw new PermissionDeniedException(Permission.DENIED, "Reporter");
		}
		ReporterData reporter = getData().getReporter();
		HibernateUtil.evictData(getDbControl().getHibernateSession(), reporter);
		return reporter;
	}
	
	/**
	 	Get the reporter in this well. 
	 	The initialization of the reporter can be forced.
	  	@param forceInitialization TRUE if the initialization of
	  		the reporter should be forced, FALSE otherwise.
	  	@return A <code>ReporterData</code> object or null if no
	  	reporter has been specified.
	  	@since 2.13
	 */
	public ReporterData getReporter(boolean forceInitialization)
	{
		if (!getSessionControl().hasPermission(Permission.READ, Item.REPORTER))
		{
			throw new PermissionDeniedException(Permission.READ, "Reporter");
		}
		ReporterData reporter = getData().getReporter();
		if (forceInitialization) org.hibernate.Hibernate.initialize(reporter);
		HibernateUtil.evictData(getDbControl().getHibernateSession(), reporter);		
		return reporter;
	}
	
	/**
		Set the reporter of this well.
		@param reporter The new reporter or null if the reporter is unknown
		@throws PermissionDeniedException If this well has been saved to 
			the database or a parent well has been set (in which case the reporter
			must always be the same as for the parent) or the logged in user
			doesn't have write permission for the well or use permission
			for reporters
	*/
	public void setReporter(ReporterData reporter)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (isInDatabase())
		{
			throw new PermissionDeniedException("Can't modify the reporter for an existing well."); 
		}
		if (getData().getParent() != null)
		{
			throw new PermissionDeniedException("Can't modify the reporter for a well with a parent."); 
		}
		getData().setReporter(reporter);
	}
}
