/**
	$Id: Settings.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.Presets;
import net.sf.basedb.core.Presets.Preset;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.DefaultFilter;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.Registry;

/**
	Class for keeping settings for the extensions system. Settings
	are kept in the <code>settings.xml</code> file in the plug-ins
	directory.
	<p>

	Extension points and extensions can be disabled/enabled by
	{@link #enableExtensionPoint(String, boolean)} and 
	{@link #enableExtension(String, boolean)}.
	<p>

	This class also implements the {@link ExtensionsFilter}
	interface, and can be used on the 
	{@link Registry#useExtensions(net.sf.basedb.util.extensions.ClientContext, ExtensionsFilter, String...)}
	method. This means that disabling/enabling extensions and extension points 
	will immediately be visible in the web interface. Sorting is delegated
	to the {@link DefaultFilter}.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class Settings 
	extends DefaultFilter
	implements ExtensionsFilter 
{
	
	private final File file;
	private final ExtensionsManager manager;
	private boolean hasChanged;
	private Presets presets;
	private Preset settings;
	private Preset installedFiles;
	private Preset disabledExtensions;
	private Preset disabledExtensionPoints;
	
	/**
		Create a new Settings object. If the given file doesn't exists
		it will be created if {@link #save()} is called. The settings file
		must be an XML file that can be read by the {@link Presets} class.
		@param file The file to load/save settings in. If null, settings only are 
			kept in memory and are lost when the server is shutdown.
	*/
	Settings(ExtensionsManager manager, File file)
	{
		this.manager = manager;
		this.file = file;
		this.presets = new Presets();
		if (file != null && file.isFile())
		{
			try
			{
				InputStream in = new BufferedInputStream(new FileInputStream(file));
				presets.loadFrom(in, file.getAbsolutePath());
				in.close();
			}
			catch (IOException ex)
			{
				throw new RuntimeException(ex);
			}
		}
		this.settings = presets.getDefault();
		this.installedFiles = presets.getPreset("installed-files");
		this.disabledExtensionPoints = presets.getPreset("disabled-extension-points");
		this.disabledExtensions = presets.getPreset("disabled-extensions");
		hasChanged = false;
	}
	
	/*
		From the ExtensionsFilter interface
		-----------------------------------
	*/
	@Override
	public boolean isEnabled(ExtensionPoint<?> extensionPoint) 
	{
		if (extensionPoint == null) return false;
		ExtensionsFile xtFile = manager.getFileByObjectKey(new ExtensionPointKey(extensionPoint));
		if (xtFile != null && (xtFile.hasError() || !xtFile.isValid())) return false;
		return disabledExtensionPoints.getSetting(extensionPoint.getId()) == null;
	}

	@Override
	public boolean isEnabled(Extension<?> extension) 
	{
		if (extension == null) return false;
		ExtensionsFile xtFile = manager.getFileByObjectKey(new ExtensionKey(extension));
		if (xtFile != null && (xtFile.hasError() || !xtFile.isValid())) return false;
		return disabledExtensions.getSetting(extension.getId()) == null;
	}
	// --------------------------------------------------

	/**
		Enable/disable an extension.
		@param extensionId The ID of the extension to enable/disable
		@param enable TRUE to enable the extension, FALSE to disable it
	*/
	public synchronized void enableExtension(String extensionId, boolean enable)
	{
		disabledExtensions.setSetting(extensionId, enable ? null : "1");
		hasChanged = true;
	}
	
	/**
		Enable/disable an extension point.
		@param extensionPointId The ID of the extension point to enable/disable
		@param enable TRUE to enable the extension point, FALSE to disable it
	*/
	public synchronized void enableExtensionPoint(String extensionPointId, boolean enable)
	{
		disabledExtensionPoints.setSetting(extensionPointId, enable ? null : "1");
		hasChanged = true;
	}
	
	/**
		Get a preset containing settings for the given extesion. If no
		settings exists, null is returned. Do not make modifications to
		the returned preset since the result is undefined. Use
		{@link #setSetting(String, String, String)} to set values.
		@since 3.4
	*/
	public Preset getSettingsForExtension(String extensionId)
	{
		return presets.exists(extensionId) ? presets.getPreset(extensionId) : null;
	}
	
	/**
		Set a configuration setting for an extension. Do not forget to
		{@link #save()}.
		@param extensionId The ID of the extension to save a setting for
		@param key The name of the setting
		@param value The value of the setting, use null to remove
		@since 3.4
	*/
	public synchronized void setSetting(String extensionId, String key, String value)
	{
		if (value == null && !presets.exists(extensionId)) return;
		
		Preset p = presets.getPreset(extensionId);
		p.setSetting(key, value);
		if (p.size() == 0) presets.deletePreset(extensionId);
		hasChanged = true;
	}

	/**
		Get a configuration setting for an extension.
		@param extensionId The ID of the extension to save a setting for
		@param key The name of the setting
		@return The value or null if it doesn't exists
		@since 3.4
	*/
	public String getSetting(String extensionId, String key)
	{
		Preset p = getSettingsForExtension(extensionId);
		return p == null ? null : p.getSetting(key);
	}
	
	/**
		Check if the given file is marked as installed.
		@since 3.0
	*/
	public boolean isInstalledFile(File file)
	{
		return Values.getBoolean(installedFiles.getSetting(file.getAbsolutePath()));
	}
	
	/**
		Mark the given file as installed by the extension system.
		@since 3.0
	*/
	public void setInstalledFile(File file)
	{
		if (file == null) return;
		hasChanged = true;
		installedFiles.setSetting(file.getAbsolutePath(), "1");
	}
	
	/**
		Mark the given file as uninstalled.
		@since 3.0
	*/
	public void setUninstalledFile(File file)
	{
		if (file == null) return;
		hasChanged = true;
		installedFiles.setSetting(file.getAbsolutePath(), null);
	}
	
	/**
		Get a list with all files that are installed.
		@since 3.0
	*/
	public List<File> getInstalledFiles()
	{
		List<String> paths = installedFiles.getKeys();
		List<File> files = new ArrayList<File>(paths.size());
		for (String path : paths)
		{
			files.add(new File(path));
		}
		return files;
	}
	
	
	/**
		Save the settings to disk. If the settings has not been
		changed the file is not written.
	*/
	public synchronized void save()
	{
		if (!hasChanged || file == null) return;
		try
		{
			OutputStream out = new FileOutputStream(file);
			presets.writeTo(out);
			out.close();
		}
		catch (IOException ex)
		{
			throw new RuntimeException(ex);
		}
		hasChanged = false;
	}
}
