/**
	$Id: VariableConverter.java 5946 2012-02-03 13:18:47Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.xml;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.RendererFactory;

/**
	A value converter that can be used on all setter methods
	that has been annotated with the {@link VariableSetter}
	annotation. The input string will be scanned for
	parts that match <code>$VARIABLE$</code> and have that part
	replaced by the actual value of a defined variable. Undefined
	variables are not replaced. See {@link #setVariable(String, String)}
	for more details.

	<p>
	Web client note! The web client application defines the variables
	<code>ROOT</code> and <code>HOME</code> which are the paths to
	the web application root directory (<code>/base2</code>) and the 
	home directory (<code>/base2/extensions/xxxx</code>) of the current 
	extension. This means that the XML configuration file may
	contain values like <code>$ROOT$/images/copy.png</code>. The 
	<code>$ROOT$</code> portion will automatically be replaced with the
	correct path if the factory method is annotated with the {@link VariableSetter}
	annotation.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class VariableConverter 
	implements ValueConverter
{

	// Pattern to match $VARIABLE$
	private static final Pattern VARIABLE = Pattern.compile("\\$(\\w+)\\$");
	
	// Contains variable values
	private Map<String, String> variables;
	
	/**
		Create a new variable converter. Use {@link #setVariable(String, String)}
		to define variables and {@link XmlLoader#addValueConverter(ValueConverter)}
		to register this converter with an XML loader.
	*/
	public VariableConverter()
	{}
	
	/*
		From the ValueConverter interface
		----------------------------------
	*/
	@Override
	public String convert(String in, Method method) 
	{
		if (variables == null) return in;
		if (!method.isAnnotationPresent(VariableSetter.class)) return in;
		Matcher m = VARIABLE.matcher(in);
		StringBuffer sb = new StringBuffer();
		while (m.find())
		{
			String variable = m.group(1);
			String value = variables.get(variable);
			if (value == null) value = m.group();
			m.appendReplacement(sb, Matcher.quoteReplacement(value));
		}
		m.appendTail(sb);
		return sb.toString();
	}
	// ---------------------------------------

	/**
		Set the value of a variable. If a setter method of a factory class 
		({@link ActionFactory} or {@link RendererFactory}) is annotated with the
		{@link VariableSetter} annotation the 
		XML loader will scan the value from the XML file for replacement 
		tags, <code>$VARIABLE$</code>, and replace the tag with the value 
		of the variable. If the variable doesn't exists, the tag is left 
		untouched.
		
		@param name The name of the variable
		@param value The value of the variable, or null to remove
			the variable
	*/
	public void setVariable(String name, String value)
	{
		if (variables == null) variables = new HashMap<String, String>();
		if (variables == null) variables = new HashMap<String, String>();
		if (value == null)
		{
			variables.remove(name);
		}
		else
		{
			variables.put(name, value);
		}		
	}
	
	
	/**
		Get the value of a variable.
		@param name The variable's name
		@return The value, or null if it doesn't exists
	*/
	public String getVariable(String name)
	{
		return variables == null ? null : variables.get(name);
	}
}
