/**
	$Id: MatrixSpotIntensityEventHandler.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.bfs.BfsParser;
import net.sf.basedb.util.bfs.DataParser;
import net.sf.basedb.util.bfs.EventHandler;
import net.sf.basedb.util.bfs.EventType;
import net.sf.basedb.util.importer.spotdata.SynchronizedSpotDataParser.SynchronizedData;

/**
	Event handler that extracts spot intensity information from matrix BFS
	data files and inserts the values to a spot batcher. In the matrix format 
	there is one data file per channel. In each file there should be one data 
	column for each assay. All data files should have the same 
	column layout and the same number of lines. For example, file1.txt 
	contains data for channel 1 and file2.txt contains data for channel 2.
	In both files the first column (index=0) has intensity for assay #1
	and the second column (index=1) has intensity for assay #2.
	<p>
	This event handler is designed to work together with a {@link SynchronizedSpotDataParser}
	and will only react to {@link SynchronizedSpotDataParser#DATA_EVENT} events.
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (må, 20 apr 2015) $
*/
public class MatrixSpotIntensityEventHandler
	implements EventHandler
{

	private final SpotBatcher spotBatcher;
	private final PositionEventHandler positionHandler;
	private final BioAssay[] assays;
	private final int[] chIndex;
	private float[] ch;
	
	/**
		Create a new spot instensity event handler.
		@param spotBatcher The spot batcher that should receive the intensity data
		@param positionHandler The event handler that is responsible for extracting
			position information from the reporter annotations file
		@param assays An array with all assays in the same order that they
			appeared in the pdata file
		@param chIndex An array with file indexes for channel intensities,
			first value is the file index for channel 1. The first file has
			index = 0. The size of this array should match the number of channels
			in the experiment.
	*/
	public MatrixSpotIntensityEventHandler(SpotBatcher spotBatcher, 
		PositionEventHandler positionHandler, BioAssay[] assays, int[] chIndex)
	{
		if (spotBatcher == null) throw new NullPointerException("spotBatcher");
		if (positionHandler == null) throw new NullPointerException("positionHandler");
		if (assays == null) throw new NullPointerException("assays");
		if (chIndex == null) throw new NullPointerException("chIndex");
		this.spotBatcher = spotBatcher;
		this.positionHandler = positionHandler;
		this.assays = assays;
		this.chIndex = chIndex;
		this.ch = new float[chIndex.length];
	}
	
	/**
		Extract the spot intensity values from the data files and batch insert
		them using the spot batcher. If all channels are empty, no data will be
		inserted for that spot, otherwise a null value will be inserted for
		the missing channels. The parser will treat invalid values as
		empty columns.
	*/
	@SuppressWarnings("rawtypes")
	@Override
	public void handleEvent(EventType eventType, Object eventData, BfsParser parser)
	{
		if (eventType == SynchronizedSpotDataParser.DATA_EVENT)
		{
			SynchronizedData[] data = (SynchronizedData[])eventData;
			int position = positionHandler.getCurrentPosition();
			for (int i = 0; i < assays.length; ++i)
			{
				boolean allNaN = true;
				for (int j = 0; j < ch.length; ++j)
				{
					SynchronizedData assayData = data[chIndex[j]+1];
					ch[j] = Values.getFloat(assayData.getData()[i], Float.NaN);
					allNaN &= Float.isNaN(ch[j]);
				}
				if (!allNaN) 
				{
					spotBatcher.insert(assays[i].getDataCubeColumnNo(), position, ch);
				}
			}
		}
		else if (eventType == DataParser.END_OF_FILE_EVENT)
		{
			spotBatcher.flush();
		}
	}

}
