/**
	$Id: $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;

public abstract class NameableNodeValidator<I extends Nameable>
	extends BasicNodeValidator<I> 
{
	/**
		Creates a new validator.
		@param ruleMissing The validation rule that is broken if an item is missing,
			or null to not report missing items as a failure
		@param ruleDenied The validation rule that is broken if we are denied access
			to an item, or null to not report this as a failure
	 */
	public NameableNodeValidator(Validator ruleMissing, Validator ruleDenied)
	{
		super(ruleMissing, ruleDenied);
	}

	@SuppressWarnings("unchecked")
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		super.postValidate(dc, context, node, parentNode);
		
		BasicItem item = node.getItem();
		String name = ((Nameable)item).getName();
		Item itemType = node.getItemType();
		Map<String, List<Node>> namedNodes = (Map<String, List<Node>>)context.getCachedObject("names.used");
		Set<String> knownDuplicates = (Set<String>)context.getCachedObject("names.duplicate");
		if (namedNodes == null)
		{
			namedNodes = new HashMap<String, List<Node>>();
			context.setCachedObject("names.used", namedNodes);
		}
		if (knownDuplicates == null)
		{
			knownDuplicates = new HashSet<String>();
			context.setCachedObject("names.duplicate", knownDuplicates);
		}
		
		// If we have seen this name before, there is an entry in the cache
		// with a list of all nodes that contains an item with the same name
		// NOTE! Multiple nodes may reference the SAME item. This should
		// of course not be reported as duplicates.
		List<Node> list = namedNodes.get(name);
		if (list != null)
		{
			for (Node other : list)
			{
				if (itemType != other.getItemType())
				{
					// An item of a different type has the same name
					createFailure(context, knownDuplicates, Validator.NONUNIQUE_NAME_GLOBAL, node, name);
					createFailure(context, knownDuplicates, Validator.NONUNIQUE_NAME_GLOBAL, other, name);
				}
				else if (!item.equals(other.getItem()))
				{
					// Another item of the same type has the same name
					createFailure(context, knownDuplicates, Validator.NONUNIQUE_NAME_FOR_TYPE, node, name);
					createFailure(context, knownDuplicates, Validator.NONUNIQUE_NAME_FOR_TYPE, other, name);
				}
			}
			list.add(node);
		}
		else
		{
			list = new LinkedList<Node>();
			namedNodes.put(name, list);
		}
		list.add(node);
	}

	/**
		Create a failure on this node unless we have created a failure 
		for the same validator already.
	*/
	private void createFailure(OverviewContext context, Set<String> knownDuplicates, 
			Validator validator, Node node, String name)
	{
		if (knownDuplicates.add(node.getId() + ":" + validator.getId()))
		{
			String itemName = node.getItemType().toString().toLowerCase();
			// This node was previously an unknown duplicate
			context.createFailure(validator, node, 
				validator.getFailureSummary() + " for " + itemName + ": '" + name + "'", 
				new Fix("Change name of " + itemName, node.getItem()));
		}
	}
	
}
