/**
	$Id: SampleImporter.java 6998 2015-11-04 07:16:05Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BioMaterial;
import net.sf.basedb.core.BioMaterialEvent;
import net.sf.basedb.core.BioMaterialEventSource;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Kit;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.DefaultItemFunction;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing sample items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2015-11-04 08:16:05 +0100 (on, 04 nov 2015) $
*/
public class SampleImporter
	extends AbstractItemImporter<Sample>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.SAMPLE, GuiContext.Type.LIST));

	/**
		Column mapping parameter for the sample creation date.
	*/
	protected static final PluginParameter<String> createdColumnMapping = 
		new PluginParameter<String>(
			"createdColumnMapping",
			"Created",
			"Mapping that picks the creation date from the data columns. The date format is " +
			"specified by the 'Date format' parameter." +
			"Example: \\Created\\",
			optionalColumnMapping
		);
	
	/**
		Column mapping parameter for the original quantity.
	*/
	protected static final PluginParameter<String> originalQuantityColumnMapping = 
		new PluginParameter<String>(
				"originalQuantityColumnMapping",
				"Original quantity",
				"Mapping that picks the original quantity (µg) from the data columns. The " +
				"values must be in the numeric format specified by the 'Number format' parameter." +
				"Example: \\Original quantity\\",
				optionalColumnMapping
		);
	
	/**
		Column mapping parameter for the parent type.
	*/
	protected static final PluginParameter<String> parentTypeColumnMapping = 
		new PluginParameter<String>(
				"parentTypeColumnMapping",
				"Parent type",
				"Mapping that picks the data column specifying the type of parent item. " +
				"Recognised values are BIOSOURCE, SAMPLE and EXTRACT, but only two of them are " +
				"valid for each importer. If the parent type is not specified, the importer will " +
				"try to guess a parent type from the item subtype. Example: \\Parent type\\",
				optionalColumnMapping
		);

	/**
		Column mapping parameter for the parent (biosource OR sample)
	*/
	protected static final PluginParameter<String> parentColumnMapping = 
		new PluginParameter<String>(
				"parentColumnMapping",
				"Parent",
				"Mapping that picks the name or id of the parent item in the data columns. " +
				"Example: \\Parent\\",
				optionalColumnMapping
		);
	
	/**
		Column mapping parameter for the used quantity from the parent.
	*/
	protected static final PluginParameter<String> usedQuantityColumnMapping = 
		new PluginParameter<String>(
				"usedQuantityColumnMapping",
				"Used quantity from parent",
				"Mapping that picks the used quantity (µg) of the parent item in the data columns. The " +
				"values must be in the numeric format specified by the 'Number format' parameter." +
				"Example: \\Used quantity\\",
				optionalColumnMapping
		);

	/**
	 	Column mapping parameter for the bioplate
	 */
	protected static final PluginParameter<String> bioPlateColumnMapping =
		new PluginParameter<String>(
				"bioPlateColumnMapping",
				"Bioplate",
				"Mapping that picks the name or id of the bioplate in the data columns. "+
				"Example: \\Bioplate\\",
				optionalColumnMapping
		);
	

	/**
	 	Column mapping parameter for the biowell's row
	 */
	protected static final PluginParameter<String> bioWellRowColumnMapping =
		new PluginParameter<String>(
				"bioWellRowColumnMapping",
				"Biowell row",
				"Mapping that picks the row of the biowell in the data columns. " +
				"The row coordinate is usually given as alphabetical letters, " +
				"A, B, C, etc, but it can also be given as a numerical value " +
				"where A=1, B=2 and so on. Note that BASE internally uses a " +
				"0-based coordinate system. The coordinate values given in the " +
				"file are converted to this system before data is saved. "+
				"Example: \\Biowell row\\",
				optionalColumnMapping
		);
	
	/**
	 	Column mapping parameter for the biowell's column
	 */
	protected static final PluginParameter<String> bioWellColColumnMapping =
		new PluginParameter<String>(
				"bioWellColColumnMapping",
				"Biowell column",
				"Mapping that picks the column of the biowell in the data columns. " +
				"The column coordinate is usually given as a number, but alphabetical " +
				"letters are also supported, A=1, B=2 and so on. Note that BASE internally uses a " +
				"0-based coordinate system. The coordinate values given in the " +
				"file are converted to this system before data is saved. "+
				"Example: \\Biowell column\\",
				optionalColumnMapping
		);
	
	public SampleImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper externalIdMapper;
	private Mapper originalQuantityMapper;
	private Mapper createdMapper;
	private Mapper protocolMapper;
	private Mapper kitMapper;
	private DefaultItemFunction<Protocol> defaultPrototolFunction;
	private Mapper parentTypeMapper;
	private Mapper parentMapper;
	private Mapper usedQuantityMapper;
	private Mapper bioPlateMapper;
	private Mapper bioWellRowMapper;
	private Mapper bioWellColMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link PropertyIdMethod#EXTERNAL_ID}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME, PropertyIdMethod.EXTERNAL_ID, 
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	@Override
	protected Item getItemForSubtypes()
	{
		return Item.SAMPLE;
	}

	/**
		Adds column mappings for name, externalId, description,
		protocol, creation date, original quantity, pooled, parent
		and used quantity from parent.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(subtypeColumnMapping);
		parameters.add(externalIdColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(protocolColumnMapping);
		parameters.add(kitColumnMapping);
		
		parameters.add(createdColumnMapping);
		parameters.add(originalQuantityColumnMapping);
		parameters.add(parentTypeColumnMapping);
		parameters.add(parentColumnMapping);
		parameters.add(usedQuantityColumnMapping);
		parameters.add(bioPlateColumnMapping);
		parameters.add(bioWellRowColumnMapping);
		parameters.add(bioWellColColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Calls {@link Sample#getQuery()}.
	*/
	@Override
	protected ItemQuery<Sample> createItemQuery()
	{
		return Sample.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? Sample.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? Sample.MAX_DESCRIPTION_LENGTH : null, null);
		externalIdMapper = getMapper(ffp, (String)job.getValue("externalIdColumnMapping"), 
				cropStrings ? Sample.MAX_EXTERNAL_ID_LENGTH : null, null);
		originalQuantityMapper = getMapper(ffp, (String)job.getValue("originalQuantityColumnMapping"), null, null);
		createdMapper = getMapper(ffp, (String)job.getValue("createdColumnMapping"), null, null);
		defaultPrototolFunction = getDefaultItemFunction(Item.PROTOCOL);
		protocolMapper = getMapper(ffp, (String)job.getValue("protocolColumnMapping"), null, null, defaultPrototolFunction);
		kitMapper = getMapper(ffp, (String)job.getValue("kitColumnMapping"), null, null);
		parentTypeMapper = getMapper(ffp, (String)job.getValue("parentTypeColumnMapping"), null, null);
		parentMapper = getMapper(ffp, (String)job.getValue("parentColumnMapping"), null, null);
		usedQuantityMapper = getMapper(ffp, (String)job.getValue("usedQuantityColumnMapping"), null, null);
		bioPlateMapper = getMapper(ffp, (String)job.getValue("bioPlateColumnMapping"), null, null);
		bioWellRowMapper = getMapper(ffp, (String)job.getValue("bioWellRowColumnMapping"), null, null);
		bioWellColMapper = getMapper(ffp, (String)job.getValue("bioWellColColumnMapping"), null, null);
	}
	
	@Override
	protected Sample createItem(DbControl dc, FlatFileParser.Data data)
	{
		Sample s = Sample.getNew(dc);
		updateItem(dc, s, data);
		return s;
	}
	
	@Override
	protected void updateItem(DbControl dc, Sample sample, FlatFileParser.Data data)
	{
		if (nameMapper != null) sample.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) sample.setDescription(descriptionMapper.getValue(data));
		if (externalIdMapper != null) sample.setExternalId(externalIdMapper.getValue(data));
		if (originalQuantityMapper != null) sample.setOriginalQuantity(originalQuantityMapper.getFloat(data));
		BioMaterialEvent creationEvent = sample.getCreationEvent();
		if (createdMapper != null) creationEvent.setEventDate(parseDate(createdMapper.getValue(data)));
		updateItemSubtype(dc, sample, data);
		if (protocolMapper != null) 
		{
			ItemSubtype type = ItemSubtype.getRelatedSubtype(dc, sample, Item.PROTOCOL, SystemItems.getId(Protocol.SAMPLING));
			if (defaultPrototolFunction != null) defaultPrototolFunction.setCurrentSubtype(type);
			String nameOrId = protocolMapper.getValue(data);
			Protocol protocol = findProtocol(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (nameOrId == null || protocol != null) creationEvent.setProtocol(protocol);
		}
		if (kitMapper != null) 
		{
			ItemSubtype type = ItemSubtype.getRelatedSubtype(dc, sample, Item.KIT, SystemItems.getId(Kit.SAMPLING));
			String nameOrId = kitMapper.getValue(data);
			Kit kit = findKit(dc, FallbackIdMethod.NAME_OR_ID, nameOrId, type);
			if (nameOrId == null || kit != null) creationEvent.setKit(kit);
		}
		if (bioPlateMapper != null && bioWellColMapper != null && bioWellRowMapper != null)
		{
			String plateNameOrId = bioPlateMapper.getValue(data);
			String row = bioWellRowMapper.getValue(data);
			String column = bioWellColMapper.getValue(data);
			BioPlate plate = findBioPlate(dc, FallbackIdMethod.NAME_OR_ID, plateNameOrId);
			BioWell well = findBioWell(dc, plate, row, column);
			if (plateNameOrId == null || row == null || column == null || well != null)
			{
				sample.setBioWell(well);
			}
		}

		if (parentMapper != null)
		{
			// Try to find a parent type
			Item parentType = sample.getParentType();
			// Is it specified in the imported file?
			if (parentTypeMapper != null)
			{
				String type = parentTypeMapper.getValue(data);
				if (type != null) parentType = Item.valueOf(type);
			}
			
			// See if we can find a related subtype
			ItemSubtype type = null;
			if (parentType == null)
			{
				// If we don't know the parent type, check for both biosurce and sample types
				type = ItemSubtype.getRelatedSubtype(dc, sample, Item.BIOSOURCE, Item.SAMPLE);
				if (type != null) parentType = type.getMainItemType();
			}
			else
			{
				type = ItemSubtype.getRelatedSubtype(dc, sample, parentType);
			}
			
			// Try to find either a sample or a biosource
			String nameOrId = parentMapper.getValue(data);
			BioMaterial parent = null;
			if (parentType == Item.SAMPLE)
			{
				parent = findSample(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			}
			else
			{
				parent = findBioSource(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			}
			if (nameOrId == null || parent != null) 
			{
				BioMaterialEventSource evtSrc = sample.getCreationEvent().setSource(parent);
				if (evtSrc != null) 
				{
					Float usedQuantity = usedQuantityMapper == null ? null : usedQuantityMapper.getFloat(data);
					evtSrc.setUsedQuantity(usedQuantity);
				}
			
			}
		}
	}
	
	/**
		If the sample is pooled, add the parent sample.
	*/
	@Override
	protected void updateMultiLineItem(DbControl dc, Sample sample, 
		FlatFileParser.Data data, int multiLineNum)
	{
		if (parentMapper != null && sample.getParentType() == Item.SAMPLE)
		{
			String nameOrId = parentMapper.getValue(data);
			ItemSubtype type = ItemSubtype.getRelatedSubtype(dc, sample, Item.SAMPLE, 0);
			Sample parent = findSample(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (parent != null)
			{
				Float usedQuantity = usedQuantityMapper == null ?
						null : usedQuantityMapper.getFloat(data);
				sample.getCreationEvent().addSource(parent).setUsedQuantity(usedQuantity);
			}
		}
	}
	// ------------------------------------
}
