/**
	$Id: PrepareTest.java 5827 2011-10-26 10:51:52Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test.performance;


import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Trashcan;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.test.FileUtil;
import net.sf.basedb.test.MouseData;
import net.sf.basedb.test.PluginUtil;
import net.sf.basedb.test.TestUtil;
import net.sf.basedb.util.ConsoleProgressReporter;

/**
	Prepares the performance test by uploading data files, importing reporters 
	and creating a single array design. If the files, reporters and array
	design already exists, this test does nothing.

	@author nicklas
	@version 2.4
	@base.modified $Date: 2011-10-26 12:51:52 +0200 (on, 26 okt 2011) $
*/
public class PrepareTest
{

	public static boolean ok = true;

	private static PluginConfiguration reporterImporter;
	private static PluginConfiguration rawDataImporter;
	private static PluginConfiguration featureImporter;
	private static ArrayDesign design;
	private static Experiment experiment;
	private static File rawData1;
	private static File rawData2;
	private static File reporters;
	
	/**
		Run the tests in this class.
	*/
	public static boolean run(String user, String password)
	{
		DbControl dc = null;
		TestUtil.write("++Entering preparation\n");
		try
		{
			TestUtil.login(user, password);
			dc = TestUtil.getDbControl();

			// We use the GENERIC platform
			Platform generic = Platform.getByExternalId(dc, Platform.GENERIC);
			
			// Create experiment
			experiment = Util.findExperiment(dc, "PerformanceTest");
			if (experiment == null)
			{
				experiment = Util.createExperiment(dc, "PerformanceTest");
			}
			
			// Create array design
			design = Util.findArrayDesign(dc, "PerformanceTest");
			if (design == null)
			{
				design = Util.createArrayDesign(dc, generic, "PerformanceTest");
			}		
			
			// Upload files
			reporters = FileUtil.uploadFile(dc, "/mouse", "plates_and_reporters.mouse.v4.37k.txt", File.REPORTER);
			rawData1 = FileUtil.uploadFile(dc, "/mouse", "genepix.mouse.v4.37k.00h.gpr", File.RAW_DATA);
			rawData2 = FileUtil.uploadFile(dc, "/mouse", "genepix.mouse.v4.37k.24h.gpr", File.RAW_DATA);
			
			// Create importers
			// Importer for reporters
			reporterImporter = PluginUtil.createPluginConfiguration(dc, 
				"PerformanceTest", 
				"net.sf.basedb.plugins.ReporterFlatFileImporter", MouseData.getReporterParameters());

			// Importer for raw data
			rawDataImporter = PluginUtil.createPluginConfiguration(dc, 
				"PerformanceTest", 
				"net.sf.basedb.plugins.RawDataFlatFileImporter", MouseData.getGenePixRawDataParameters(false));

			// Importer for features
			featureImporter = PluginUtil.createPluginConfiguration(dc, 
				"PerformanceTest",
				"net.sf.basedb.plugins.ReporterMapFlatFileImporter", MouseData.getGenePixReporterMapParameters());
			dc.commit();
				
			
			// Import features
			if (!design.hasFeatures()) 
			{
				// Import reporters
				PluginUtil.importReporters(reporterImporter, reporters, "PerformanceTest: Import reporters");
				importFeatures(featureImporter, rawData1, design);
			}
			
			TestUtil.logout();
		}
		catch (Throwable t)
		{
			ok = false;
			t.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		TestUtil.write("++Leaving preparation (" + (ok ? "OK" : "FAILED") + ")\n");
		return ok;
	}

	/**
		Import features.
	*/
	public static void importFeatures(PluginConfiguration plugin, File file, ArrayDesign design)
		throws Exception
	{
		TestUtil.write("--Importing features from file: " + file.getName() + "\n");
		DbControl dc = TestUtil.getDbControl();
		try
		{
			plugin = PluginConfiguration.getById(dc, plugin.getId());
			Job job = Job.getNew(dc, plugin.getPluginDefinition(), plugin, null);
			job.setName("PerformanceTest: Import feature to " + design.getName());
			job.setParameterValue("file", new FileParameterType(), file);
			job.setParameterValue("arrayDesign", 
				new ItemParameterType<ArrayDesign>(ArrayDesign.class, null), design);
			job.setParameterValue("missingReporterError", new StringParameterType(), "skip");
			dc.saveItem(job);
			dc.commit();
			PluginUtil.executeJob(job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	public static boolean cleanUp(String user, String password)
	{
		DbControl dc = null;
		TestUtil.write("++Entering cleanup test\n");
		try
		{
			TestUtil.login(user, password);

			List<BasicItem> itemsToRemove = new ArrayList<BasicItem>();
			itemsToRemove.add(experiment);
			itemsToRemove.add(design);
			itemsToRemove.add(rawData1);
			itemsToRemove.add(rawData2);
			itemsToRemove.add(reporters);
			itemsToRemove.add(reporterImporter);
			itemsToRemove.add(rawDataImporter);
			itemsToRemove.add(featureImporter);
			
			dc = TestUtil.getDbControl();
			ItemQuery<Job> jobQuery = Job.getQuery();
			jobQuery.restrict(Restrictions.like(Hql.property("name"), Expressions.string("PerformanceTest%")));
			itemsToRemove.addAll(jobQuery.list(dc));
			ItemQuery<RawBioAssay> query = experiment.getRawBioAssays();
			itemsToRemove.addAll(query.list(dc));
			dc.commit();
			
			ProgressReporter progress = new ConsoleProgressReporter(true);
			int numDeleted = Trashcan.delete(TestUtil.getSessionControl(), itemsToRemove, true, progress);
			progress.display(100, "Deleted " + numDeleted + " of " + itemsToRemove.size() + " items\n");
			
			TestUtil.logout();
		}
		catch (Throwable t)
		{
			ok = false;
			t.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		TestUtil.write("++Leaving cleanup test (" + (ok ? "OK" : "FAILED") + ")\n");
		return ok;
	}
	

	public static Experiment getExperiment(DbControl dc)
	{
		return Experiment.getById(dc, experiment.getId());
	}

	public static ArrayDesign getArrayDesign(DbControl dc)
	{
		return ArrayDesign.getById(dc, design.getId());
	}
	
	public static PluginConfiguration getRawDataImporter(DbControl dc)
	{
		return PluginConfiguration.getById(dc, rawDataImporter.getId());
	}
	
	public static File getRawDataFile1(DbControl dc)
	{
		return File.getById(dc, rawData1.getId());
	}

	public static File getRawDataFile2(DbControl dc)
	{
		return File.getById(dc, rawData2.getId());
	}
}
